/*****************************************************************************/
/* Kernel SOCKS Bouncer (Loadable Kernel Module) for 2.6.x kernels [KSB26]   */
/* (c) 2004 Paolo Ardoino <ardoino.gnu@disi.unige.it>                        */
/*****************************************************************************/
/*									     */
/* This program is free software; you can redistribute it and/or modify	     */
/* it under the terms of the GNU General Public License as published by	     */
/* the Free Software Foundation; either version 2 of the License, or	     */
/* (at your option) any later version.					     */
/* This program is distributed in the hope that it will be useful,	     */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of	     */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	     */
/* GNU General Public License for more details.				     */
/*									     */
/* You should have received a copy of the GNU General Public License	     */
/* along with this program; if not, write to the Free Software		     */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*****************************************************************************/

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/moduleparam.h>
#include <asm/unistd.h>
#include <asm/uaccess.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/fs.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <asm/semaphore.h>
#include "include/ksb26.h"

#define KSB26_DEV_NAME "/dev/ksb26"
#define KSB26_IOCTL_BUFLEN 4096
#define KSB26_MAXSOCKS 2

#define KSB26_INIT_LIST_HEAD(ptr) do { \
	(ptr)->next = (ptr); (ptr)->prev = (ptr); \
} while(0)
#define KSB26_CLEAR_HSTRUCT(ksb26_hs) do { \
	memset(ksb26_hs.ip, '\0', 18); \
	ksb26_hs.port = 0; \
	ksb26_hs.naddr = 0; \
	ksb26_hs.nport = 0; \
	ksb26_hs.wrk = 0; } while(0)

#define DEBUG 1

extern char ksb26;
extern unsigned int ksb26_nsocks;
extern struct ksb26_host ksb26_slh;
extern struct ksb26_host ksb26_bhlh;
extern struct semaphore ksb26_listsem;
extern struct semaphore ksb26_cntsem;
extern int ksb26_connect_cnt;
extern int ksb26_maxsocks;

static int nsocks = KSB26_MAXSOCKS;
static int ksb26_major;
static char ksb26_ioctl_msg[KSB26_IOCTL_BUFLEN];
static int ksb26_dev_bool = 0;

static int device_open(struct inode *, struct file *);
static int device_release(struct inode *, struct file *);
static ssize_t device_write(struct file *, const char *, size_t, loff_t *);

static int ksb26_isip(char *ip)
{
	int pcnt = 0;
	char *ptr;

	for(ptr = ip; *ptr != '\0'; ptr++) {
		if(*ptr == '.')
			pcnt++;
	}
	if(pcnt == 3)
		return 1;
	return 0;
}

static int ksb26_istcpport(int port) {
	if(port >= 1 && port <= 65535)
		return 1;
	return 0;
}

static int ksb26_getline(char *string, int init_ch, int end_ch, char **line)
{
	char *aux = NULL;
	int ccnt = 0;
  
	*line = NULL;
	if (!string)
		return 0;
	if(init_ch != -1) {
		for(; *string != init_ch && *string != '\0' && *string != end_ch; ++string);
		string++;
	}
	aux = string;
	for ( ; *aux != end_ch && *aux != '\0'; aux++, ccnt++);
	*line = kmalloc((int)(aux - string + 1) * sizeof(char), GFP_KERNEL);
	snprintf(*line,(int)(aux - string + 1), "%s", string);
	return ccnt;
}

static int ksb26_atoi(char *str)
{
	int i, n = 0;
	
	for(i = 0; str[i] < '0' || str[i] > '9'; ++i);
	for(n = 0; str[i] != '\0' && str[i] >= '0' && str[i] <= '9'; ++i)
		n = 10 * n + (str[i] - '0');
	return n;
}

static void ksb26_list_add(struct ksb26_host *new, struct ksb26_host *prev, struct ksb26_host *next)
{
	next->prev = new;
	new->next = next;
	new->prev = prev;
	prev->next = new;
}

static void ksb26_list_add_tail(struct ksb26_host *new, struct ksb26_host *head)
{
	ksb26_list_add(new, head->prev, head);
}

static void ksb26_list_del(struct ksb26_host *prev, struct ksb26_host *next)
{	
	next->prev = prev;
	prev->next = next;
}

static struct file_operations fops = {
	.write = device_write,
	.open = device_open,
	.release = device_release
};

static int device_open(struct inode *inode, struct file *file)
{
	if(ksb26_dev_bool == 1) 
		return -EBUSY;	
	ksb26_dev_bool++;
	try_module_get(THIS_MODULE);
	return 0;
}

static int device_release(struct inode *inode, struct file *file)
{
	ksb26_dev_bool--;
	module_put(THIS_MODULE);
	return 0;
}

static ssize_t device_write(struct file *filp, const char *buffer, size_t lenght, loff_t *offset)
{
	int bytes_write = 0, linelen = 0;
	char *aux, *line;
	char *tbuf;
	int tport = 0;
	struct ksb26_host *ksb26_haux;
	
	for(bytes_write = 0; bytes_write < lenght && bytes_write < KSB26_IOCTL_BUFLEN; bytes_write++)
		get_user(ksb26_ioctl_msg[bytes_write], buffer + bytes_write);
	ksb26_ioctl_msg[lenght - 1] = '\0';
	aux = ksb26_ioctl_msg;
	down_interruptible(&ksb26_listsem);
	while(aux < ksb26_ioctl_msg + lenght) {
		if((linelen = ksb26_getline(aux, -1, '\n', &line)) == 0)
			break;
		if(line[0] == 'S') {
			if(ksb26_getline(line, ':', ';', &tbuf) > 0) {
				tport = ksb26_atoi(tbuf);
				kfree(tbuf);
				if(ksb26_getline(line, 'S', ':', &tbuf) > 0) {
					if(ksb26_isip(tbuf) && ksb26_istcpport(tport)) {
						ksb26_haux = kmalloc(sizeof(struct ksb26_host), GFP_KERNEL);
						ksb26_list_add_tail(ksb26_haux, &ksb26_slh);
						strncpy(ksb26_slh.prev->ip, tbuf, 18);
						ksb26_slh.prev->port = tport;
						ksb26_slh.prev->naddr = in_aton(ksb26_slh.prev->ip);
						ksb26_slh.prev->nport = htons(ksb26_slh.prev->port);
						ksb26_slh.prev->wrk = 1;
						ksb26_haux = NULL;
						ksb26_nsocks++;
						printk("[%s] %s:%d socks added.\n", MODNAME, ksb26_slh.prev->ip, ksb26_slh.prev->port);
					}
					kfree(tbuf);
				}
			}
		} else if(line[0] == 'H') {
			if(ksb26_getline(line, 'H', ';', &tbuf) > 0) {
				if(ksb26_isip(tbuf)) {
					ksb26_haux = kmalloc(sizeof(struct ksb26_host), GFP_KERNEL);
					ksb26_list_add_tail(ksb26_haux, &ksb26_bhlh);
					strncpy(ksb26_bhlh.prev->ip, tbuf, 18);
					ksb26_bhlh.prev->naddr = in_aton(ksb26_bhlh.prev->ip);
					ksb26_haux = NULL;
					printk("[%s] %s target host added.\n", MODNAME, ksb26_bhlh.prev->ip);
				}
				kfree(tbuf);
			}
		} else {
			if(strncmp(line, "csl", 3) == 0) {
				while((ksb26_haux = ksb26_slh.prev) != &ksb26_slh) {
					ksb26_nsocks++;
					ksb26_list_del(ksb26_haux->prev, &ksb26_slh);
					kfree(ksb26_haux);
				}
				ksb26_nsocks = 0;
			}
		}
		aux = aux + linelen + 1;
		kfree(line);
	}
	up(&ksb26_listsem);
	return bytes_write;
}

static int __init modinit(void)
{
	printk("[%s v%s] MODULE LOADED\n", MODNAME, KSB26_VERSION);
	if((ksb26_major = register_chrdev(0, KSB26_DEV_NAME, &fops)) < 0) {
		printk("[%s] Cannot open device %s with %d major.\n", MODNAME, KSB26_DEV_NAME, ksb26_major);
		return -ENODEV;
	}
	printk("[%s] Device %s opened.\n", MODNAME, KSB26_DEV_NAME);
	printk("[%s] Major device number = %d.\n", MODNAME, ksb26_major);
	printk("[%s] mknod %s c %d 0\n", MODNAME, KSB26_DEV_NAME, ksb26_major);
	KSB26_INIT_LIST_HEAD(&ksb26_slh);
	KSB26_CLEAR_HSTRUCT(ksb26_slh);
	KSB26_INIT_LIST_HEAD(&ksb26_bhlh);
	KSB26_CLEAR_HSTRUCT(ksb26_bhlh);
	ksb26 = 1;
	ksb26_connect_cnt = 0;
	ksb26_nsocks = 0;
	sema_init(&ksb26_listsem, 1);
	sema_init(&ksb26_cntsem, 1);
	printk("[%s] Hijacking SYS_CONNECT syscall.\n", MODNAME);
	ksb26_maxsocks = nsocks;
	printk("[%s] ksb26_maxsocks = %d\n", MODNAME, ksb26_maxsocks);
	return 0;
}

static void __exit modcleanup(void)
{
	ksb26 = 0;
	if(unregister_chrdev(ksb26_major, KSB26_DEV_NAME) < 0)
		printk("[%s] Cannot unregister device %s.\n", MODNAME, KSB26_DEV_NAME);
	printk("[%s] MODULE CLEANED\n", MODNAME);
	return;
}

module_init(modinit);
module_exit(modcleanup);

module_param(nsocks, int, 0);

MODULE_AUTHOR("Paolo Ardoino");
MODULE_DESCRIPTION("Kernel SOCKS Bouncer for 2.6.x kernels\n");
MODULE_LICENSE("GPL");
